# Author: Claude Anderson:  November, 2008

class MinHeap:
    """ Implements an indirect heap so it can support 
      the Isin and Decrease operations that are not supported efficiently by 
      an ordinary binary heap."""
    
    def __init__(self, key):
        """key: list of values from which we build initial heap"""
        self.n = len(key)-1
        self.key = key
        self.into = [i for i in range(self.n + 1)]
        self.outof = [i for i in range(self.n + 1)]
        self.heapify()
    
    def heapify(self):
        for i in range(self.n//2, 0, -1):
            self.siftdown(i, self.n)
            
    def siftdown(self, i, n):
        """ sift down for a minHeap.
        i is the heap index,(not the index into the key array)"""
        s = self.outof[i]
        temp = self.key[s]
        while 2*i <= n:
            c = 2*i   # c is for child
            if c < n and self.key[self.outof[c+1]] <  \
                         self.key[self.outof[c]]:
                c += 1
            if self.key[self.outof[c]] < temp:
                self.outof[i] = self.outof[c]
                self.into[self.outof[i]] = i
            else:
                break
            i = c
            self.outof[i] = s
            self.into[s] = i
                
            
    
    def __str__(self):
        return "n: " + str(self.n) + "\n" + \
                "key:   " + str(self.key) + "\n" + \
               "into:  " + str(self.into) + "\n" + \
               "outof: " + str(self.outof) 
    
    
    def delete(self):
        """delete the mimimum value from this heap, returning its value"""
        result = self.outof[1]
        temp = self.outof[1]
        self.outof[1] = self.outof[self.n]
        self.into[self.outof[1]] = 1
        self.outof[self.n] = temp
        self.into[temp] = self.n
        self.n -= 1
        self.siftdown(1, self.n)
        return result
        
    
    def isIn(self, w):
        """ returns True iff w is in this heap """
        return self.into[w] <= self.n
    
    def keyVal(self, w):
        """ returns the weight corresponding to w"""
        return self.key[w]
    
    def decrease(self, w, newWeight):
        """ change the weight corresponding to 
        vertex w  to newWeight (which must be no 
        larger than its current weight) """
        # p is for parent, c is for child
        self.key[w] = newWeight
        c = self.into[w]
        p = c//2
        while p >= 1:
            if self.key[self.outof[p]] <= newWeight:
                break
            self.outof[c] = self.outof[p]
            self.into[self.outof[c]] = c
            c = p
            p = c//2
        self.outof[c] = w
        self.into[w] = c
    
 
def main():  
    keys = [False, 66, 12, 312, 25, 8 ,109, 7, 18]
    heap = MinHeap(keys)
    print (heap)
    print ("keyVal(5): ",  heap.keyVal(5))
    print ("deleting:", heap.delete())
    print (heap)

    print ("deleting:", heap.delete())
    print (heap)
    print (heap.isIn(5), heap.isIn(2))
    heap.decrease(4, 3)
    print ("decreasing key 4 to 3")
    print (heap)
    
    
if __name__ == "__main__":
    main()